#! /usr/bin/python

"""
The addressbook holds a list of permission slips for recipients to which we
are allowed to send mail, and incoming permissions for senders from whom we
will accept mail.
"""

from petmail import util, permission, transport, crypto, pbutil
from petmail.interfaces import IAddressBook, IIDRecord, IAddressBookEntry
from twisted.python.components import implements
from twisted.internet import defer

import os

class AddressBook(object):
    # TODO: needs test coverage
    __implements__ = IAddressBook
    debug = 0

    def __init__(self, agent):
        self.agent = agent
        self.entries = []

    def notifyWatchers(self):
        self.agent.addressBookChanged(self)

    def lookupID(self, ID):
        assert(isinstance(ID, type("")))
        for e in self.entries:
            if e.idRecord and e.idRecord.ID == ID:
                return e
        return None
    def lookupPetName(self, name):
        for e in self.entries:
            if e.petname == name:
                return e
        return None
    def lookupAddress(self, address):
        for e in self.entries:
            if address in e.idRecord.addresses:
                return e

    def addEntry(self, entry, message=None):
        # message is the message which caused this entry to be added
        assert(isinstance(entry, AddressBookEntry))
        self.entries.append(entry)
        self.notifyWatchers()

    def addRecord(self, record):
        e = self.lookupID(record.ID)
        if e:
            # TODO: a bit silly, as we verify/decode twice
            e.idRecord.update(record.signed, self.agent.crypto)
            if self.debug:
                print "addRecord", record.genericPermission
            self.notifyWatchers()
            return
        e = AddressBookEntry(self)
        e.idRecord = record
        self.entries.append(e)
        self.notifyWatchers()
        return e


class AddressBookEntry(object):
    __implements__ = IAddressBookEntry
    def __init__(self, addressbook):
        self.addressbook = addressbook
        self.petname = None
        self.idRecord = None
        self.ourPermission = None   # what we use to send mail to them
        self.theirPermission = None # what we require when they send to us
        self.lastSentOwnerEntry = 0 # the last time we sent them our IDRecord
        self.lastSentTheirPermission = 0 # last time we sent .theirPermission

    def setPetName(self, name):
        self.petname = name

    def update(self, encoded):
        """Update this entry with the encoded IDRecord they just sent us.
        """
        if not self.idRecord:
            self.idRecord = IDRecord()
        self.idRecord.update(encoded, self.addressbook.agent.crypto)
        self.lastUpdated = util.now()

    def updateOurPermission(self, permstring):
        # TODO: permstring will eventually be signed
        self.ourPermission = permission.decode(permstring)

    def getName(self):
        if self.petname != None:
            return self.petname
        if self.idRecord and self.idRecord.name != None:
            return '"%s"' % self.idRecord.name
        if self.idRecord:
            return self.idRecord.ID
        return "????"

    def getTransport(self, agent, message):
        """Obtain a Transport object which can be used to deliver an
        encapsulated message to this recipient.
        """
        t = transport.transportFromString(self.idRecord.transport, agent)
        return t

class IDRecord(object):
    """This record describes a recipient. Each recipient in the agent's
    addressbook has one. Each is uniquely identified by the hash of the
    public key.

    The owner of the agent has some OwnerEntry records, and these are
    published in various ways. When somebody looks up one of those published
    records, the data they retrieve is turned into an IDRecord.
    
    Serialized forms of these records are retrieved in the process of
    sending a message. This is the form in which they are stored in the
    addressbook and the 'me' records.
    """

    __implements__ = IIDRecord

    def __init__(self, signed=None, ca=None):
        self.ID = None     # hash of the public key
        self.pubkey = None # public key
        # identifying information (all optional)
        self.name = ""
        self.addresses = [] # email addresses which point to this record
        # permission record
        self.genericPermission = None # non-sender-specific permissions
        # transport record
        self.transportString = None
        # other stuff
        self.expires = None
        self.signed = None
        if signed:
            self.update(signed, ca)

    def update(self, signed, ca):
        """Take an encoded form (the output of OwnerEntry.encode) and
        populate this record."""
        sigID, plaintext = ca.verify(signed)
        d = util.decode(plaintext)
        if d['ID'] != sigID:
            raise ValueError("claimed ID:%s, signature ID:%s" % \
                             (d['ID'], sigID))
        pubkey = ca.addPubkey(d['pubkey'])
        if pubkey.getID() != d['ID']:
            raise ValueError("pubkey ID:%s, claimed ID:%s" % \
                             (pubkey.getID(), d['ID']))

        if self.pubkey and self.pubkey.getID() != d['ID']:
            # TODO: handle key changes
            raise NotImplementedError("pubkey update")
        if self.ID and self.ID != d['ID']:
            raise NotImplementedError("pubkey update")
        self.pubkey = pubkey
        self.ID = d['ID']
        self.name = d['name']
        self.addresses = d['addresses']
        self.expires = d['expires']
        slip = d.get('permission')
        if slip:
            if not self.genericPermission:
                self.genericPermission = permission.PermissionSlip()
            self.genericPermission.update(slip, ca)
        transports = d['transports']
        self.transportString = None
        if transports:
            self.transportString = transports[0] # TODO
        self.signed = signed # to send to others

    def getTransport(self, agent, message):
        tport = transport.transportFromString(self.transportString, agent)
        return defer.succeed(tport)

class FakeNameServer(object):
    def __init__(self):
        self.names = {}
        self.IDs = {}
    def publish(self, name, record):
        signed = record.signed()
        self.names[name] = signed
        self.IDs[record.ID] = signed
    def lookup(self, agent, name):
        return IDRecord(self.names.get(name), agent.crypto)
    def lookupID(self, agent, ID):
        return IDRecord(self.IDs.get(ID), agent.crypto)
    def describe(self):
        return "Fake"

class DiskNameServer(object):
    def __init__(self, args):
        self.mapdir = args

    def lookup(self, agent, address):
        """Take a name and return a (single) ID Record, using a server of
        some sort."""
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        for f in os.listdir(mapdir):
            try:
                s = open(os.path.join(mapdir, f), "r").read()
                r = IDRecord(s, agent.crypto)
            except Exception, e:
                print "lookup: couldn't check file '%s'" % address
                print e
                continue
            if address in r.addresses:
                return r
        return None

    def lookupID(self, agent, ID):
        """Take an ID number and return a (single) ID Record, using a server
        of some sort."""
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        for f in os.listdir(mapdir):
            try:
                s = open(os.path.join(mapdir, f), "r").read()
                r = IDRecord(s, agent.crypto)
            except Exception, e:
                print "lookupID: couldn't check file '%s'" % address
                print e
                continue
            if ID == r.ID:
                return r
        return None

    def publish(self, address, record):
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        f = open(os.path.join(mapdir, record.ID), "w")
        f.write(record.signed())
        f.close()

    def describe(self):
        return "Disk:%s" % self.mapdir

class PBNameServer(object):
    def __init__(self, url):
        self.hostname, portnum = url.split(":")
        self.portnum = int(portnum)

    def lookup(self, agent, address):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._lookup2, "lookup", agent, address)
        return d

    def lookupID(self, agent, ID):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._lookup2, "lookupID", agent, ID)
        return d

    def _lookup2(self, server, method, agent, address):
        d = server.callRemote(method, address)
        d.addCallback(self._lookup3, agent)
        return d

    def _lookup3(self, s, agent):
        r = IDRecord(s, agent.crypto)
        return r

    def publish(self, address, record):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._publish2, address, record)
        return d

    def _publish2(self, server, address, record):
        d = server.callRemote("publish", record.signed())
        return d

    def describe(self):
        return "PB:%s:%d" % (self.hostname, self.portnum)    

def getLookupAgentFromString(args):
    scheme, rest = args.split(":", 1)
    fac = {"Fake": FakeNameServer,
           "Disk": DiskNameServer,
           "PB": PBNameServer,
           }
    return fac[scheme](rest)
