#! /usr/bin/python

from StringIO import StringIO

from twisted.python.components import implements
from twisted.internet import defer
from twisted.python import failure, log

from petmail import util, crypto, interfaces
NotPermitted = interfaces.NotPermitted

class InvalidMessage(Exception):
    """The message was mangled somehow"""

class BaseMessage(object):
    debug = False
    def __init__(self, agent):
        self.agent = agent
        # delivery attributes
        self.when = util.now()  # the nominal creation time of the message
        self.eventlog = [] # log of what processors have done
        self.doDump = False
        self.provides = []
        # recipientEntry, recipientRecord, grantedPermission
        # senderEntry, acceptedRFC822
        self.senderEntry = None

    def startProcessing(self):
        if self.agent.debugProcessing:
            print "startProcessing"
        self.processors = self.getProcessors()
        self.processing = []
        self.done = False
        self.failed = False
        d = self.processingDeferred = defer.Deferred()
        self.process()
        return d
        
    def process(self):
        """Do whatever can be done to the message."""
        # this function will not raise Unsendable exceptions. Instead, it
        # (or a callback) will run self.processingFailed
        debug = self.agent.debugProcessing
        if debug:
            print "process:"
            print " processors %s" % self.processors
            print " provides %s" % self.provides
            print " processing %s" % self.processing

        if not self.processors and not self.processing:
            if debug:
                print " all done"
            return self.processingComplete()
        runnable = self.findRunnableProcesses()
        if not runnable and not self.processing:
            if debug:
                print " no processors are runnable"
            f = failure.Failure
            why = f(interfaces.Unsendable("no processors are runnable"))
            return self.processingFailed(why)

        if debug and not runnable:
            print " weird, nothing to do"
        self.processing.extend(runnable)
        for p in runnable:
            d = defer.maybeDeferred(p.checkMessage, self)
            if self.debug: print " %s gave %s" % (p, d)
            d.addCallback(self.processorComplete, p)
            d.addErrback(self.processorFailed, p)
        if debug:
            print " done"

    def findRunnableProcesses(self):
        # which processors can be run? Removes them from self.processors
        runnable = []
        need = []
        for p in self.processors[:]:
            canRun = True
            for d in p.getDependencies():
                if not d in self.provides:
                    if d not in need: need.append(d)
                    canRun = False
                    break
            needToRun = False
            # empty provides means it must always be run
            processorProvides = p.getProvides()
            if not processorProvides:
                needToRun = True
            else:
                for f in processorProvides:
                    if not f in self.provides:
                        needToRun = True
            if not needToRun:
                if self.agent.debugProcessing:
                    print "don't need to run", p
                self.processors.remove(p)
            elif canRun:
                self.processors.remove(p)
                runnable.append(p)
        if self.agent.debugProcessing:
            print " runnable:", runnable
        return runnable

    def processorComplete(self, res, processor):
        if self.agent.debugProcessing:
            print "processorComplete", processor
        self.processing.remove(processor)
        if not self.failed:
            self.process()
    def processorFailed(self, why, processor):
        if self.agent.debugProcessing:
            print "processorFailed", processor, why
        self.processing.remove(processor)
        self.processingFailed(why)

    def processingComplete(self):
        # this should only be called once, when the last processor succeeds.
        assert not self.done
        self.done = True
        self.processingDeferred.callback(self)
    def processingFailed(self, why):
        # this may be called multiple times, once per failing processor
        log.msg("processing step failed:", why)
        if self.failed:
            if self.agent.debugProcessing:
                print "processingFailed >1 times"
            return
        self.done = self.failed = True
        self.processingDeferred.errback(why)

    def provide(self, which):
        if which not in self.provides:
            self.provides.append(which)

    def hasNode(self, name):
        return self.headers.has_key(name)
    def getNode(self, name, default=None):
        return self.headers.get(name, default)

    def log(self, message):
        self.eventlog.append(message)


class Message(BaseMessage):
    """This object represents a single message being sent to some individual
    recipient. Its life-cycle is as follows:

     1: Created when the user hits the 'Send' button on their MUA,
        containing an rfc822-formatted header+body pair.
     2: The recipient's AddressBookEntry is looked up.
     3: .encode() is called with the AddressBookEntry. This returns a string
        which should be given to the transport layer for delivery to the
        recipient.
     4: The Message is stored in a list of outstanding messages (indexed by
        messageID). An ack from the far end will remove it from the list. If
        it stays on the list for too long (5 days), the user is notified
        that the message did not make it through.
    """
    __implements__ = interfaces.IMessage

    def __init__(self, agent):
        BaseMessage.__init__(self, agent)
        self.messageID = util.generateMessageID()
        self.recipientAddress = None
        self.recipientRecord = None # 'recipientRecord'
        self.recipientEntry = None # 'recipientEntry'
        self.transport = None # the transport used to send the message
        self.sent = None # the time this message was sent. TODO: set it
        self.headers = {} # indexed by name

    def setMessageRecipient(self, name):
        # this accepts pet names and addresses
        r = self.agent.addressbook.lookupPetName(name)
        if not r:
            r = self.agent.addressbook.lookupAddress(name)
        if r:
            self.setRecipientEntry(r)
        else:
            self.recipientAddress = name # will be looked up later

    def setRecipientRecord(self, r):
        assert implements(r, interfaces.IIDRecord)
        self.recipientRecord = r
        self.provide('recipientRecord')

    def setRecipientEntry(self, e):
        assert implements(e, interfaces.IAddressBookEntry)
        self.recipientEntry = e
        self.provide('recipientEntry')
        self.setRecipientRecord(e.idRecord)

    def setSenderEntry(self, e):
        assert implements(e, interfaces.IOwnerEntry)
        self.senderEntry = e
        self.provide('senderEntry')

    def getProcessors(self):
        return self.agent.getOutboundProcessors()

    def addNode(self, name, value):
        assert (name not in self.headers)
        self.headers[name] = value

    def addRFC822(self, body): #, us, recipient, body, addrbook=None):
        """Add an RFC822 message section to the Message."""

        # construct the header
        header = {}
        #header['bodyHash'] = util.hash(body)
        #header['bodyLen'] = len(body)
        ##header['senderID'] = us.ID
        ##header['messageID'] = util.generateMessageID()
        ##header['timestamp'] = self.when
        header['tickets'] = []
        header['body'] = body

        self.addNode("rfc822", header)

    def send(self):
        self.agent.sendMessage(self)

    def addTicket(self, ticket):
        # ticket is a signed blob
        assert(self.headers['rfc822'])
        self.headers['rfc822']['tickets'].append(ticket)

    def encodeHeaders(self):
        return util.encode(self.headers)

    def deliver(self):
        enc = self.encrypt(self.senderEntry.privkey,
                           self.recipientRecord.pubkey)
        self.transport.send(enc)

    def encrypt(self, us, recipient):
        if not isinstance(us, crypto.PrivateKey):
            print "us", us
        assert isinstance(us, crypto.PrivateKey)
        assert isinstance(recipient, crypto.PublicKey)

        encoded = util.encode(self.headers)
        return recipient.encryptAndSign(us, encoded)

    def dump(self):
        if not self.doDump:
            return
        keys = self.headers.keys()
        keys.sort()
        print ",".join(keys)
        for m in self.eventlog:
            print " %s" % m


class ReceivedMessage(BaseMessage):
    """A ReceivedMessage object is created for each incoming message. The
    .process method is run on it to decrypt and check permissions. Once this
    succeeds, the following attributes will be available on the
    ReceivedMessage object:

    @type senderID:  C{string}
    @ivar senderID:  the ID of the message sender
    @type sender:    C{AddressBookEntry}
    @ivar sender:    our entry for this sender (maybe None for new senders??)
    @type body:      C{string}
    @ivar body:      the body of the message

    TODO: add the permission with which the message was accepted
    """

    __implements__ = interfaces.IReceivedMessage

    def __init__(self, agent, encrypted):
        """If any part of the message is invalid, this will raise
        InvalidMessage or UndecryptableError. If the sender does not have
        permission, then it will raise NotPermitted."""

        BaseMessage.__init__(self, agent)

        if isinstance(encrypted, type("")):
            encrypted = StringIO(encrypted)
        self.encrypted = encrypted
        # .senderEntry
        # .acceptedRFC822
        # .senderID: filled in by agent after decryption
        # .sigID: filled in by .decrypt  #TODO: get rid of one of these

    def decrypt(self):
        # TODO: must let crypto layer tell us which recipient was used
        #recipient = self.agent.me.privkey
        ca = self.agent.crypto
        recipient, sigID, plaintext = ca.decrypt(self.encrypted)
        #self.signer, decrypted = recipient.decrypt(self.encrypted)
        self.headers = util.decode(plaintext)
        assert recipient.getID() == self.agent.me.privkey.getID() # for now
        return sigID

    def getProcessors(self):
        return self.agent.getInboundProcessors()

    def dump(self):
        if not self.doDump:
            return
        print "Message processed"
        if self.senderEntry:
            print " From:", self.senderEntry.getName()
        if self.acceptedRFC822:
            l = len(self.getNode('rfc822')['body'])
            print " RFC822 len(%d) accepted" % l
        for m in self.eventlog:
            print " %s" % m
        
