#! /usr/bin/python

from twisted.cred import credentials
from twisted.spread import pb
from twisted.internet import reactor, defer
from twisted.python import log
import sys, re, os
from twisted.internet.protocol import ProcessProtocol
from twisted.trial import unittest
from twisted.python.components import implements

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import addressbook, permission, agent, interfaces, message, \
     owner, crypto

class DebugChildPP(ProcessProtocol):
    verbose = 0
    def __init__(self, d):
        self.waiting = 1
        self.buffer = ""
        self.d = d
    def connectionMade(self):
        if self.verbose: print "server seen"
    def outReceived(self, data):
        if self.verbose: print "server[stdout]:", data
        if self.verbose: print "endserver[stdout]"
        if not self.waiting:
            return
        self.buffer += data
        r = re.search(r'^debugService: (\d+)$', self.buffer)
        if r:
            self.gotPortNum(int(r.group(1)))
            self.waiting = 0
    def gotPortNum(self, portnum):
        self.d.callback(portnum)
    def errReceived(self, data):
        print "STDERR:", data
    def processEnded(self, status):
        if self.verbose: print "server ended", status

class Catcher(pb.Referenceable):
    def remote_newMessage(self, body):
        print "newMessage", body
    def remote_transport(self, encapsulated):
        #print "transport", encapsulated
        self.test.results = encapsulated

class AgentProgram:
    verbose = 1
    def __init__(self, test, argv):
        self.test = test
        self.argv = argv
    def launch(self):
        d = defer.Deferred()
        self.pp = DebugChildPP(d)
        p = reactor.spawnProcess(self.pp,
                                 self.argv[0], self.argv,
                                 os.environ, ".")
        if self.verbose: print "starting server.."
        self.process = p
        return d
    def connect(self, portnum):
        if self.verbose: print "connecting to %d" % portnum
        self.f = pb.PBClientFactory()
        d = self.f.login(credentials.UsernamePassword("debug", "pw"))
        d.addCallback(self.loggedIn)
        reactor.connectTCP("localhost", portnum, self.f)
        return d
    def loggedIn(self, perspective):
        self.remote = perspective
        d = self.remote.callRemote("test", "yar")
        d.addCallbacks(lambda res, self=self: self.setup(),
                       self.err, errbackArgs=["test"])
        return d
    def setup(self):
        if self.verbose: print "doing setup"
        self.catcher = Catcher()
        self.catcher.test = self.test
        d = self.remote.callRemote("registerNewMessageHandler",
                                   self.catcher)
        d.addErrback(self.err, "registerNewMessageHandler")
        d.addCallback(lambda res, self=self:
                      self.remote.callRemote("registerTransport",
                                             self.catcher))
        d.addErrback(self.err, "registerTransport")
        return d
    def kill(self):
        self.process.signalProcess("INT")
    def err(self, why, where=None):
        if where:
            print "error in:", where
        log.err(why)
        self.kill()
        return why

class ClientTest(unittest.TestCase):
    skip = "remote agent control is in flux"
    def setup(self):
        self.results = None

    def setup1(self, res):
        self.call = self.prog.remote.callRemote
        print "setupClient1"
        d = self.call("init")
        d.addErrback(self.prog.err, "init")
        d.addCallback(self.setup2)
        return d
    def setup2(self, ID):
        self.ID = ID
        d = self.call("lookupPetName", ID, "alice")
        d.addErrback(self.prog.err, "lookupPetName")
        d.addCallback(self.gotAlice)
        return d
    def gotAlice(self, aliceID):
        self.failUnless(aliceID)
        body = \
"""From: bob@boo.far
To: alice@foo.bar
Subject: Yo

Testing
 -Bob
"""
        d = self.call("sendMessage", self.ID, [aliceID], body)
        d.addErrback(self.prog.err, "sendMessage")
        d.addCallback(self.sentMessage)
        return d
    def sentMessage(self, res):
        # the message should be sent by the time this function completes
        self.failUnless(self.results)
        return 1
    def testClient1(self):
        p = AgentProgram(self, ["../petmail/agent.py"])
        self.prog = p
        d = p.launch().addCallback(p.connect)
        d.addCallback(self.setup1)
        d.addErrback(p.err)
        r = unittest.deferredResult(d)
        self.failUnless(r)
        p.kill()

class FakeAgent(agent.Agent):
    def acceptRFC822(self, rm, why):
        self.inbox.append((rm,why))



class NonLiveTest(unittest.TestCase):

    def build(self):
        server = addressbook.FakeNameServer()

        self.crypto = crypto.EncryptionAgent()

        alice_agent = FakeAgent('aliceAgent')
        alice_agent.populateProcessors()
        alice_agent.crypto = self.crypto
        alice_agent.addressbook = addressbook.AddressBook(alice_agent)
        alice = owner.OwnerEntry()
        alice.setServiceParent(alice_agent)
        alice.publishers = [server]
        alice.setKey(self.crypto.generateKey("alice"))
        alice.setName("Alice")
        alice.addAddress("alice@foo.com")
        alice.setGenericPermission(permission.PermissionSlip())
        alice.setInlet("Dummy:alice")
        alice.publish()
        alice_agent.me = alice
        alice_agent.registerAddressLookupAgent(server)

        bob_agent = FakeAgent("bobAgent")
        bob_agent.populateProcessors()
        bob_agent.crypto = self.crypto
        bob_agent.addressbook = addressbook.AddressBook(bob_agent)
        bob = owner.OwnerEntry()
        bob.setServiceParent(bob_agent)
        bob.publishers = [server]
        bob.setKey(self.crypto.generateKey("bob"))
        bob.setName("Bob")
        bob.addAddress("bob@foo.com")
        bob.setGenericPermission(permission.PermissionSlip())
        bob.setInlet("Dummy:bob")
        bob.publish()
        bob_agent.me = bob
        bob_agent.registerAddressLookupAgent(server)

        return (alice_agent, alice), (bob_agent, bob), server

    def test1(self):
        (alice_agent, alice), (bob_agent, bob), server = self.build()
        d = alice_agent.retrieveAddress("bob@foo.com")
        bob_record = unittest.deferredResult(d)
        self.failUnless(isinstance(bob_record, addressbook.IDRecord))

        body = \
"""From: alice@foo.com
To: bob@foo.com
Subject: Yar

Testing
 -Alice
"""
        bob_inbox = []
        bob_agent.inbox = bob_inbox
        #alice_agent.sendMessage0(alice, bob_record, body)
        m = message.Message(alice_agent)
        m.setSenderEntry(alice)
        m.setRecipientRecord(bob_record)
        #m.setRecipientEntry(None)
        m.recipientEntry = None; m.provide('recipientEntry')
        m.addRFC822(body)
        sent = []
        def deliver(msg, s=sent):
            s.append(msg)
            #print "deliver", msg
            #print "transport", msg.transport
        #alice_agent.debugProcessing = True

        d = m.startProcessing()
        sent = unittest.deferredResult(d)
        self.failUnlessIdentical(sent, m)
        self.failUnless(sent.getNode("messageID"))
        self.failUnless(sent.getNode("rfc822"))

    def test2(self):
        (alice_agent, alice), (bob_agent, bob), server = self.build()
        d = alice_agent.retrieveAddress("bob@foo.com")
        bob_record = unittest.deferredResult(d)
        self.failUnless(isinstance(bob_record, addressbook.IDRecord))

        body = \
"""From: alice@foo.com
To: bob@foo.com
Subject: Yar

Testing
 -Alice
"""
        bob_inbox = []
        bob_agent.inbox = bob_inbox
        m = message.Message(alice_agent)
        m.setSenderEntry(alice)
        m.setRecipientRecord(bob_record)
        #m.setRecipientEntry(None)
        m.recipientEntry = None; m.provide('recipientEntry')
        m.addRFC822(body)

        #alice_agent.debugProcessing = True
        #bob_agent.debugProcessing = True

        m.send()

        self.failUnlessEqual(len(bob_inbox), 1)
        inmsg,why = bob_inbox[0]

        self.failUnless(implements(inmsg, interfaces.IReceivedMessage))
        self.failUnlessEqual(inmsg.headers['rfc822']['body'], body)
        self.failUnlessEqual(why, "(generic) no ticket required")

if __name__ == '__main__':
    c = Client(int(sys.argv[1]))
    reactor.run()
