#! /usr/bin/python

from StringIO import StringIO

from twisted.trial import unittest

from petmail import util, crypto

# the real key generation setup is expensive (10-20 seconds), so only do it
# once. When using the resulting agents, be careful to not modify them in any
# way (like adding pubkeys) that will influence later tests.

def setup(cryptoClass):
    agent = cryptoClass("signtest")

    priv1 = agent.generateKey("key1")
    keyid1 = priv1.getID()
    pub1 = crypto.PublicKey(agent, keyid1)

    priv2 = agent.generateKey("key2")
    keyid2 = priv2.getID()
    pub2 = crypto.PublicKey(agent, keyid2)

    return agent, priv1, pub1, priv2, pub2

gnupgStuff = None
def setupGnuPG():
    # TODO: pay attention to how this gets cleaned up (if at all)
    global gnupgStuff
    if not gnupgStuff:
        gnupgStuff = setup(crypto.GnuPGEncryptionAgent)
    return gnupgStuff

fakeStuff = None
def setupFake():
    global fakeStuff
    if not fakeStuff:
        fakeStuff = setup(crypto.FakeEncryptionAgent)
    return fakeStuff

class GnuPGCryptoTest(unittest.TestCase):
    cryptoClass = crypto.GnuPGEncryptionAgent

    def setUp(self):
        self.agent, self.priv1, self.pub1, self.priv2, self.pub2 = \
                    setupGnuPG()
        
    def testExport(self):
        agent2 = self.cryptoClass("export2")

        key = self.priv1.getKey()
        newkey = agent2.addPubkey(key)
        self.failUnlessEqual(self.priv1.getID(), newkey.getID())

        # test unchanged-key-import
        newkey = agent2.addPubkey(key)
        self.failUnlessEqual(self.priv1.getID(), newkey.getID())


    def testSign1(self):
        plaintext = "as plain as the nose on your face"
        signed = self.priv1.sign(plaintext)
        (sigID, plaintext2) = self.agent.verify(signed)
        self.failUnlessEqual(sigID, self.priv1.getID())
        self.failUnlessEqual(plaintext, plaintext2)

    def testSign2(self):
        d = {'Foo': "blah", 'Bar': "bloo boo"}
        plaintext = util.encode(d)
        signed = self.priv1.sign(plaintext)
        (sigID, p2) = self.agent.verify(signed)
        self.failUnlessEqual(sigID, self.priv1.getID())
        d2 = util.decode(p2)
        self.failUnlessEqual(d, d2)


    def testEncrypt1(self):
        d = {'Foo': "blah", 'Bar': "bloo boo"}
        plaintext = util.encode(d)
        encrypted = self.pub1.encryptAndSign(self.priv2, plaintext)
        (recip, sigID, p2) = self.agent.decrypt(encrypted)
        self.failUnlessEqual(recip.getID(), self.priv1.getID())
        self.failUnlessEqual(sigID, self.priv2.getID())
        d2 = util.decode(p2)
        self.failUnlessEqual(d, d2)

    def testEncrypt2(self):
        # crypto.GnuPGEncryptionAgent doesn't currently support StringIO as
        # input
        if not self.agent.fake:
            return
        d = {'Foo': "blah", 'Bar': "bloo boo"}
        plaintext = util.encode(d)
        encrypted = self.pub1.encryptAndSign(self.priv2, plaintext)
        sio = StringIO(encrypted)
        (recip, sigID, p2) = self.agent.decrypt(sio)
        self.failUnlessEqual(recip.getID(), self.priv1.getID())
        self.failUnlessEqual(sigID, self.priv2.getID())
        d2 = util.decode(p2)
        self.failUnlessEqual(d, d2)

class FakeCryptoTest(GnuPGCryptoTest):
    cryptoClass = crypto.FakeEncryptionAgent

    def setUp(self):
        self.agent, self.priv1, self.pub1, self.priv2, self.pub2 = \
                    setupFake()
