#! /usr/bin/python

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import util, addressbook, permission, ticket, crypto, transport
from petmail.owner import OwnerEntry
from twisted.trial import unittest

class TestOwnerEntry(unittest.TestCase):
    def setUp(self):
        self.crypto = crypto.EncryptionAgent()

    def build1(self):
        e = OwnerEntry()
        e.publishers = [addressbook.FakeNameServer()]
        e.setKey(self.crypto.generateKey("alice"))
        e.setName("alice")
        e.addAddress("alice@foo.com")
        e.setGenericPermission(None) # no new people
        e.setInlet("Dummy:alice")
        e.addInlet("Dummy:alice2")
        return e

    def build2(self):
        e = self.build1()
        p0 = permission.PermissionSlip()
        p0.minIMG = 3600
        p0.maxsize = 10**6
        tt_privkey = self.crypto.generateKey("ticketserver.com")
        tt_pubkey = tt_privkey.pubkey()
        v = ticket.TicketTaker("http://ticketserver.com", None, tt_pubkey)
        p0.addTicketTaker(v)
        e.setGenericPermission(p0)
        return e

    def build3(self):
        e = self.build2()
        i0 = transport.PBInlet("localhost:9999", None)
        e.removeInlet(e.inlets[0])
        e.setInlet(i0)
        return e

    # OwnerEntries are serialized into strings, and then unserialized into
    # IDRecords.
    def testEncode1(self):
        e = self.build1()
        r0 = e.signed()

        sigID, d = self.crypto.verify(r0)
        r1 = util.decode(d)
        self.failUnlessEqual(sigID, r1['ID'])
        self.failUnlessEqual(e.privkey.getKey(), r1['pubkey'])
        self.failUnlessEqual(r1['name'], "alice")
        self.failUnlessEqual(r1['addresses'], ["alice@foo.com"])
        self.assertApproximates(r1['expires'], util.now()+120, 60)
        self.failUnlessEqual(r1['transports'], ["Dummy:alice",
                                                "Dummy:alice2"])

    def testEncode2(self):
        e = self.build2()
        r0 = e.encode()

        r1 = util.decode(r0)

        sigID, ppt = self.crypto.verify(r1['permission'])
        self.failUnlessEqual(sigID, r1['ID'])
        gp = util.decode(ppt)
        self.failUnlessEqual(gp['IMG'], 3600)
        self.failUnlessEqual(gp['maxsize'], 10**6)
        self.failUnlessEqual(gp['contentTypes'], ["*/*"])
        self.failUnlessEqual(gp['canGrantPermission'], 1)

    def testEncode3(self):
        e = self.build3()
        r0 = e.signed()

        sigID, d = self.crypto.verify(r0)
        r1 = util.decode(d)

        self.failUnlessEqual(r1['transports'], ["PB:localhost:9999"])


    def testDecode1(self):
        r0 = self.build1()
        e = r0.signed()

        r1 = addressbook.IDRecord(e, self.crypto)
        self.failUnless(isinstance(r1, addressbook.IDRecord))
        self.failUnlessEqual(r0.ID, r1.ID)
        self.failUnlessEqual(r0.privkey.getID(), r1.pubkey.getID())
        self.failUnlessEqual(r0.name, r1.name)
