#! /usr/bin/python

from twisted.trial import unittest
from twisted.internet import defer

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import agent, processors, message, addressbook, permission
from petmail import ticket, ticketserver, crypto, interfaces, util, owner
Unsendable = interfaces.Unsendable
NotPermitted = interfaces.NotPermitted

class FakeAddressBook:
    def addEntry(self, entry, message):
        self.entries.append((entry, message))
    def lookupID(self, ID):
        return None

class FakeAgent:
    debugProcessing = False

    def __init__(self):
        self.ticketAgents = [FakeTicketAgent()]

    def getInboundProcessors(self):
        return self.inboundProcessors[:]

    def getOutboundProcessors(self):
        return self.outboundProcessors[:]

    def acceptRFC822(self, message, why):
        self.rfc822.append((True, message, why))

    def rejectRFC822(self, message, why):
        assert isinstance(why, Exception) # or maybe a Failure?
        self.rfc822.append((False, message))

    def addEntry(self, e, msg):
        self.createdEntries.append(e)

    def retrieveAddress(self, address):
        for method in self.addressLookupAgents:
            d = method.lookup(self, address)
            if isinstance(d, defer.Deferred):
                return d
            if d:
                return defer.succeed(d)
        else:
            raise interfaces.Unsendable, \
                  "none of my lookup agents could handle address '%s'" % \
                  address

    def getTicketAgents(self):
        return self.ticketAgents

class FakeTicketAgent:
    def __init__(self):
        self.requests = []
    def handleRequirements(self, requirements, msg):
        dl = []
        for r in requirements:
            if r[0] == "ticket":
                d = defer.Deferred()
                self.requests.append(d)
                dl.append(d)
        return dl
    def fire(self, ticket):
        r = self.requests.pop(0)
        r.callback(ticket)
    def error(self, why):
        r = self.requests.pop(0)
        r.errback(why)

class FakeProcessor(processors.OutboundProcessor):
    dependencies = []
    provides = []
    def __init__(self, name, test):
        processors.OutboundProcessor.__init__(self)
        self.name = name
        self.test = test
    def checkMessage(self, message):
        self.test.calls.append(self.name)
        for p in self.provides:
            message.provide(p)
    def __repr__(self):
        return "<FakeProcessor %s>" % self.name

class SlowProcessor(FakeProcessor):
    def checkMessage(self, message):
        self.test.called.append(self.name)
        self.d = defer.Deferred()
        return self.d
    def callback(self, res, msg):
        self.test.completed.append(self.name)
        for p in self.provides:
            msg.provide(p)
        self.d.callback(res)
    def errback(self, why, msg):
        self.test.completed.append(self.name)
        for p in self.provides:
            msg.provide(p)
        self.d.errback(why)

class FakeLookupAgent(object):
    def __init__(self, test):
        self.test = test
    def lookup(self, agent, address):
        self.test.lookups.append((agent, address))
        self.d = defer.Deferred()
        return self.d
    def callback(self, record):
        self.d.callback(record)
    def errback(self, why):
        self.d.errback(why)

class FailingLookupAgent(FakeLookupAgent):
    def lookup(self, agent, address):
        return None

class FakeID(object):
    __implements__ = interfaces.IIDRecord
    def __init__(self, ID, test):
        self.ID = ID
        self.test = test
        self.transport = None
        self.error = False
    def getTransport(self, agent, message):
        self.test.gettransports.append((agent, message))
        if self.error:
            raise Unsendable
        if self.transport:
            return defer.succeed(self.transport)
        self.d = defer.Deferred()
        return self.d
    def callback(self, transport):
        self.d.callback(transport)
    def errback(self, why):
        self.d.errback(why)

class FakeOwnerEntry(object):
    __implements__ = interfaces.IOwnerEntry
    def __init__(self, ID, test):
        self.ID = ID
        self.test = test

class AgentTest(unittest.TestCase):
    def setup1(self):
        self.agent = agent.Agent("testagent")
        p1 = FakeProcessor("p1", self)
        p1.dependencies = []; p1.provides = ["1"]
        p2 = FakeProcessor("p2", self)
        p2.dependencies = ["4"]; p2.provides = ["2"]
        p3 = FakeProcessor("p3", self)
        p3.dependencies = ["1", "2"]; p3.provides = ["3"]
        p4 = FakeProcessor("p4", self)
        p4.dependencies = []; p4.provides = ["4"]
        self.agent.outboundProcessors = [p1, p2, p3, p4]

    def testNoProcessors1(self):
        self.setup1()
        self.agent.outboundProcessors = []
        m = message.Message(self.agent)
        m.processors = []
        self.failUnlessEqual(m.findRunnableProcesses(), [])

    def testNoProcessors2(self):
        self.setup1()
        self.agent.outboundProcessors = []
        m = message.Message(self.agent)
        d = m.startProcessing()
        dr = unittest.deferredResult(d)
        self.failUnlessEqual(dr, m)


    def runnable(self, reachable, p):
        for f in p.dependencies:
            if not f in reachable:
                return False
        return True

    def checkDependencies(self, processors, axioms):
        allDeps = []
        for p in processors:
            for f in p.dependencies:
                if f not in allDeps:
                    allDeps.append(f)
        reachable = []
        remaining = allDeps[:]
        for f in axioms:
            reachable.append(f)
            remaining.remove(f)
        while remaining:
            #print "processors", processors
            progress = False
            for p in processors[:]:
                if self.runnable(reachable, p):
                    #print " runnable", p
                    progress = True
                    for f in p.provides:
                        if f not in reachable:
                            reachable.append(f)
                        if f in remaining:
                            remaining.remove(f)
                    processors.remove(p)
            if not progress:
                why = "unrunnable Processors: %s\n" % \
                      [p.__class__.__name__ for p in processors]
                why += "unfulfilled dependencies: %s\n" % remaining
                self.fail(why)

    def testInboundDependencies(self):
        a = agent.Agent("testagent")
        a.populateProcessors()
        processors = a.getInboundProcessors()
        axioms = []
        # interestingly, we must either start with acceptedRFC822 or
        # senderEntry (CreateSenderEntry will provide senderEntry if we
        # accepted an rfc822 message). If neither is possible, most of the
        # inbound processors are unrunnable.
        self.checkDependencies(processors, axioms)

    def testOutboundDependencies(self):
        a = agent.Agent("testagent")
        a.populateProcessors()
        processors = a.getOutboundProcessors()
        axioms = ['recipientRecord', 'recipientEntry', 'senderEntry']
        self.checkDependencies(processors, axioms)


    def testAgent1(self):
        self.setup1()
        m = message.Message(self.agent)
        self.calls = []
        #self.agent.debugProcessing = True

        # interally the loop should do p1+p4, then p2, then p3
        # but it is no longer easy to test that

        d = m.startProcessing()
        r = unittest.deferredResult(d)
        self.failUnlessEqual(r, m)
        self.failUnlessEqual(self.calls, ["p1", "p4", "p2", "p3"])

    def testAgent2(self):
        self.setup1()
        m = message.Message(self.agent)
        self.calls = []

        # internally the loop should do p4, then p2, then p3
        # but it is no longer easy to test that
        m.provides = ["1"]
        d = m.startProcessing() # should skip p1, do p4, then p2, then p3
        r = unittest.deferredResult(d)
        self.failUnlessEqual(r, m)
        self.failUnlessEqual(self.calls, ["p4", "p2", "p3"])

    def setup2(self):
        self.agent = agent.Agent("testagent")
        p1 = SlowProcessor("p1", self)
        p1.dependencies = []; p1.provides = ["1"]
        p2 = SlowProcessor("p2", self)
        p2.dependencies = ["4"]; p2.provides = ["2"]
        p3 = SlowProcessor("p3", self)
        p3.dependencies = ["1", "2"]; p3.provides = ["3"]
        p4 = SlowProcessor("p4", self)
        p4.dependencies = []; p4.provides = ["4"]
        self.agent.outboundProcessors = [p1, p2, p3, p4]
        return p1, p2, p3, p4

    def testAgent3(self):
        p1, p2, p3, p4 = self.setup2()
        m = message.Message(self.agent)
        self.called = []; self.completed = []

        d = m.startProcessing()

        # on the first pass, p1 and p4 should start
        self.failUnlessEqual(self.called, ["p1", "p4"])
        self.failUnlessEqual(self.completed, [])

        p1.callback(None, m)
        self.failUnlessEqual(self.called, ["p1", "p4"])
        self.failUnlessEqual(self.completed, ["p1"])

        p4.callback(None, m)
        # calling p4 should trigger p2
        self.failUnlessEqual(self.called, ["p1", "p4", "p2"])
        self.failUnlessEqual(self.completed, ["p1", "p4"])

        p2.callback(None, m)
        # calling p2 should trigger p3
        self.failUnlessEqual(self.called, ["p1", "p4", "p2", "p3"])
        self.failUnlessEqual(self.completed, ["p1", "p4", "p2"])

        p3.callback(None, m)
        # calling p3 should complete the message
        self.failUnlessEqual(self.called, ["p1", "p4", "p2", "p3"])
        self.failUnlessEqual(self.completed, ["p1", "p4", "p2", "p3"])

        r = unittest.deferredResult(d)
        self.failUnlessEqual(r, m)

    def testAgentGetsAllProcessors(self):
        a = agent.Agent("testagent")
        a.populateProcessors()
        self.failUnlessEqual(len(a.inboundProcessors),
                             len(processors.allInbound))
        self.failUnless(isinstance(a.inboundProcessors[0],
                                   processors.allInbound[0]))
        self.failUnlessEqual(len(a.outboundProcessors),
                             len(processors.allOutbound))
        self.failUnless(isinstance(a.outboundProcessors[0],
                                   processors.allOutbound[0]))
        

class OutboundProcessorTest(unittest.TestCase):
    def setUp(self):
        self.agent = FakeAgent()
        self.agent.crypto = crypto.EncryptionAgent()
        self.agent.addressbook = FakeAddressBook()
        self.agent.addressbook.entries = []

    def testDelivery(self):
        m = message.Message(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        p = processors.Delivery()

        self.failIf(m.hasNode("messageID"))
        r = p.checkMessage(m)
        self.failUnlessEqual(r, True)
        self.failUnless(m.hasNode("messageID"))

    def testAddressLookup(self):
        la1 = FakeLookupAgent(self)
        la2 = FailingLookupAgent(self)
        self.agent.addressLookupAgents = [la1, la2]
        p = processors.AddressLookup()

        if False: # no longer relevant
            self.lookups = []
            m = message.Message(self.agent)
            m.recipientRecord = 1
            r = p.checkMessage(m)
            self.failUnless('recipientRecord' in m.provides)
            self.failUnlessEqual(r, True)
            self.failUnlessEqual(self.lookups, [])

        # make sure the LookupAgent is called correctly
        self.lookups = []
        m = message.Message(self.agent)
        m.recipientAddress = "petmail@example.com"
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(m.recipientRecord, None)
        self.failUnless('recipientRecord' not in m.provides)
        self.failUnlessEqual(self.lookups,
                             [(self.agent, "petmail@example.com")])
        # .. and that its response is inserted as the recipientRecord
        la1.d.callback(2)
        self.failUnlessEqual(m.recipientRecord, 2)
        self.failUnless('recipientRecord' in m.provides)
        dr = unittest.deferredResult(r)
        self.failUnlessEqual(dr, True)

        # check that a LookupAgent failure marks the message as Unsendable
        self.lookups = []
        m = message.Message(self.agent)
        m.recipientAddress = "petmail@example.com"
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(m.recipientRecord, None)
        self.failUnless('recipientRecord' not in m.provides)
        self.failUnlessEqual(self.lookups,
                             [(self.agent, "petmail@example.com")])
        la1.d.errback(Unsendable("oops"))
        self.failUnlessEqual(m.recipientRecord, None)
        self.failUnless('recipientRecord' in m.provides)
        e = unittest.deferredError(r)
        self.failUnless(e.check(Unsendable))

    def testRecipientEntry(self):
        # make sure that we create an address book entry for new recipients
        # when we send them an rfc822 message

        p = processors.RecipientEntry()

        # no rfc822 node, so we shouldn't create an entry
        m = message.Message(self.agent)
        rec = addressbook.IDRecord()
        m.setRecipientRecord(rec)
        self.failUnless('recipientRecord' in m.provides)
        self.failIf('recipientEntry' in m.provides)
        r = p.checkMessage(m)
        self.failUnlessEqual(r, True)
        self.failUnless('recipientEntry' in m.provides)
        self.failIf(m.recipientEntry) # should be None
        self.failIf(self.agent.addressbook.entries)

        # add rfc822 node, so we should create an entry
        m = message.Message(self.agent)
        m.addRFC822("Message Body")
        rec = addressbook.IDRecord()
        m.setRecipientRecord(rec)
        self.failUnless('recipientRecord' in m.provides)
        self.failIf('recipientEntry' in m.provides)
        r = p.checkMessage(m)
        self.failUnlessEqual(r, True)
        self.failUnless('recipientEntry' in m.provides)
        self.failUnless(m.recipientEntry)
        entry = m.recipientEntry
        self.failUnlessEqual(entry.addressbook, self.agent.addressbook)
        self.failUnlessEqual(entry.petname, None)
        self.failUnlessIdentical(entry.idRecord, rec)
        self.failUnlessEqual(entry.theirPermission, None)
        self.failUnlessEqual(entry.lastSentOwnerEntry, 0)
        self.failUnlessEqual(entry.lastSentTheirPermission, 0)
        self.failUnlessEqual(self.agent.addressbook.entries, [(entry, m)])

    def testTransport1(self):
        # immediate return
        m = message.Message(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        r = FakeID(2, self)
        r.transport = 3
        m.setRecipientRecord(r)
        self.gettransports = []
        p = processors.Transport()

        self.failIf(m.transport)
        r = p.checkMessage(m)
        dr = unittest.deferredResult(r)
        self.failUnlessEqual(dr, True)
        self.failUnlessEqual(self.gettransports, [(self.agent, m)])
        self.failUnlessEqual(m.transport, 3)

    def testTransport2(self):
        # immediate failure
        m = message.Message(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        r = FakeID(2, self)
        r.error = True
        m.setRecipientRecord(r)
        self.gettransports = []
        p = processors.Transport()

        self.failIf(m.transport)
        self.failUnlessRaises(Unsendable, p.checkMessage, m)

    def testTransport3(self):
        # deferred return
        m = message.Message(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        m.setRecipientRecord(FakeID(2, self))
        self.gettransports = []
        p = processors.Transport()

        self.failIf(m.transport)
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(self.gettransports, [(self.agent, m)])
        self.failIf(m.transport)
        
        m.recipientRecord.callback(3)
        self.failUnlessEqual(m.transport, 3)
        dr = unittest.deferredResult(r)
        self.failUnlessEqual(dr, True)

    def testTransport4(self):
        # deferred failure
        m = message.Message(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        m.setRecipientRecord(FakeID(2, self))
        self.gettransports = []
        p = processors.Transport()

        self.failIf(m.transport)
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(self.gettransports, [(self.agent, m)])
        self.failIf(m.transport)
        
        m.recipientRecord.errback(Unsendable("oops"))
        self.failUnlessEqual(m.transport, None)
        dr = unittest.deferredError(r)
        self.failUnless(dr.check(Unsendable))

    def testUpdateTheirIDRecord(self):
        raise unittest.SkipTest("not yet written")

    def testGrantReturnPermission(self):
        raise unittest.SkipTest("not yet written")

    def testUpdateTheirPermission(self):
        raise unittest.SkipTest("not yet written")

    def testRFC822_1(self):
        # simple test of a no-permission case
        m = message.Message(self.agent)
        m.addRFC822("This is the message body.")
        rr = m.recipientRecord = addressbook.IDRecord()
        # rr.genericPermission is None: no new senders are accepted
        m.provide('recipientRecord')
        m.provide('recipientEntry') # but it is None: no relationship, so no
                                    # specific permission

        p = processors.RFC822Permission()
        #p.checkMessage(m)
        self.failUnlessRaises(Unsendable, p.checkMessage, m)

    def testRFC822_2(self):
        # single ticket retrieved successfully
        ta = self.agent.ticketAgents[0]
        m = message.Message(self.agent)
        m.addRFC822("This is the message body.")
        rr = addressbook.IDRecord()
        gp = rr.genericPermission = permission.PermissionSlip()
        gp.ticketURLs = ["http://ticketserver.com"]
        # new senders require a ticket
        m.recipientRecord = rr
        m.provide('recipientRecord')
        m.provide('recipientEntry') # but it is None: no relationship, so no
                                    # specific permission

        p = processors.RFC822Permission()
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(len(ta.requests), 1)
        self.failIf(m.getNode('rfc822').get("tickets"))

        ta.fire(123)
        self.failUnlessEqual(m.getNode('rfc822').get("tickets"), [123])
        dr = unittest.deferredResult(r)
        self.failUnless(dr)

    def testRFC822_3(self):
        # ticket failed
        ta = self.agent.ticketAgents[0]
        m = message.Message(self.agent)
        m.addRFC822("This is the message body.")
        rr = addressbook.IDRecord()
        gp = rr.genericPermission = permission.PermissionSlip()
        gp.ticketURLs = ["http://ticketserver.com"]
        # new senders require a ticket
        m.recipientRecord = rr
        m.provide('recipientRecord')
        m.provide('recipientEntry') # but it is None, and no relationship
                                    # means no specific permission

        p = processors.RFC822Permission()
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(len(ta.requests), 1)
        self.failIf(m.getNode('rfc822').get("tickets"))

        ta.error(Unsendable("ticket request failed"))
        self.failIf(m.getNode('rfc822').get("tickets"))
        dr = unittest.deferredError(r)
        self.failUnless(dr.check(type(())))

class InboundProcessorTest(unittest.TestCase):
    def setUp(self):
        self.agent = FakeAgent()
        self.agent.crypto = self.crypto = crypto.EncryptionAgent()
        self.agent.me = owner.OwnerEntry()
        self.addressbook = addressbook.AddressBook(self.agent)

    def makeReceivedMessage(self):
        rm = message.ReceivedMessage(self.agent, None)
        rm.provide('senderEntry')
        rm.senderEntry = addressbook.AddressBookEntry(self.addressbook)
        rm.senderEntry.idRecord = addressbook.IDRecord()
        rm.provide("updatedIDRecord")
        rm.senderID = rm.senderEntry.idRecord.ID
        rm.sigID = rm.senderID
        rm.provide('senderEntry')
        rm.headers = {'senderID': rm.senderID,
                      'messageID': "2",
                      'rfc822': {'body': "message body", 'tickets': []},
                      }
        return rm

    def testRFC822_1(self):
        # no permission
        rm = self.makeReceivedMessage()
        p = processors.CheckRFC822Permission()
        self.agent.rfc822 = []
        p.checkMessage(rm)
        self.failUnlessEqual(self.agent.rfc822, [(False, rm)])

    def testRFC822_2(self):
        # generic permission accepts anybody
        self.agent.me.genericPermission = permission.PermissionSlip()
        rm = self.makeReceivedMessage()
        p = processors.CheckRFC822Permission()
        self.agent.rfc822 = []
        p.checkMessage(rm)
        self.failUnlessEqual(self.agent.rfc822,
                             [(True, rm, "(generic) no ticket required")])

    def testRFC822_3(self):
        # require ticket, but don't provide one
        gp = self.agent.me.genericPermission = permission.PermissionSlip()
        url = "http://ticketserver.com"
        vendkey = self.crypto.generateKey("ticketserver.com")
        tt = ticket.TicketTaker(url, self.agent, vendkey)
        gp.addTicketTaker(tt)
        
        rm = self.makeReceivedMessage()
        
        p = processors.CheckRFC822Permission()
        self.agent.rfc822 = []
        p.checkMessage(rm)
        self.failUnlessEqual(self.agent.rfc822, [(False, rm)])

    def testRFC822_4(self):
        # require ticket
        gp = self.agent.me.genericPermission = permission.PermissionSlip()
        url = "http://ticketserver.com"
        vendkey = self.crypto.generateKey("ticketserver.com")
        tt = ticket.TicketTaker(url, self.agent, vendkey)
        gp.addTicketTaker(tt)
        
        rm = self.makeReceivedMessage()
        t1 = ticketserver.Ticket(url, util.now(), "1").signed(vendkey)
        rm.headers['rfc822']['tickets'].append(t1)
        
        p = processors.CheckRFC822Permission()
        self.agent.rfc822 = []
        p.checkMessage(rm)
        self.failUnlessEqual(self.agent.rfc822,
                             [(True, rm,
                               "(generic) valid ticket from http://ticketserver.com")])

    def testRFC822_5(self):
        # specific permission allows everything

        rm = self.makeReceivedMessage()
        rm.senderEntry.theirPermission = permission.PermissionSlip()

        p = processors.CheckRFC822Permission()
        self.agent.rfc822 = []
        p.checkMessage(rm)
        self.failUnlessEqual(self.agent.rfc822,
                             [(True, rm, "(specific) no ticket required")])

    def testRFC822_6(self):
        # specific permission requires ticket, but none provided
        sp = permission.PermissionSlip()
        url = "http://ticketserver.com"
        vendkey = self.crypto.generateKey("ticketserver.com")
        tt = ticket.TicketTaker(url, self.agent, vendkey)
        sp.addTicketTaker(tt)
        
        rm = self.makeReceivedMessage()
        rm.senderEntry.theirPermission = sp
        
        p = processors.CheckRFC822Permission()
        self.agent.rfc822 = []
        p.checkMessage(rm)
        self.failUnlessEqual(self.agent.rfc822, [(False, rm)])

    def testRFC822_7(self):
        # specific permission requires ticket
        sp = permission.PermissionSlip()
        url = "http://ticketserver.com"
        vendkey = self.crypto.generateKey("ticketserver.com")
        tt = ticket.TicketTaker(url, self.agent, vendkey)
        sp.addTicketTaker(tt)
        
        rm = self.makeReceivedMessage()
        rm.senderEntry.theirPermission = sp
        t1 = ticketserver.Ticket(url, util.now(), "1").signed(vendkey)
        rm.headers['rfc822']['tickets'].append(t1)
        
        p = processors.CheckRFC822Permission()
        self.agent.rfc822 = []
        p.checkMessage(rm)
        self.failUnlessEqual(self.agent.rfc822,
                             [(True, rm,
                               "(specific) valid ticket from http://ticketserver.com")])

    def testCreateSenderEntry(self):
        rm = self.makeReceivedMessage()
        rm.acceptedRFC822 = True
        rm.provide('acceptedRFC822')
        rm.senderEntry.idRecord = addressbook.IDRecord()
        rm.provide('updatedIDRecord')

        p = processors.CreateSenderEntry()
        self.agent.createdEntries = []
        self.agent.addressbook = FakeAddressBook()
        p.checkMessage(rm)
        self.failUnlessEqual(self.agent.createdEntries, [rm.senderEntry])
        self.failUnless(rm.senderEntry)
